/*
 * Copyright (c) 2018 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>

#include <ofi_util.h>

/* Caller should hold cmap->lock */
static void util_cmap_set_key(struct util_cmap_handle *handle)
{
	handle->key = ofi_idx2key(&handle->cmap->key_idx,
		ofi_idx_insert(&handle->cmap->handles_idx, handle));
}

/* Caller should hold cmap->lock */
static void util_cmap_clear_key(struct util_cmap_handle *handle)
{
	int index = ofi_key2idx(&handle->cmap->key_idx, handle->key);

	if (!ofi_idx_is_valid(&handle->cmap->handles_idx, index))
		FI_WARN(handle->cmap->av->prov, FI_LOG_AV, "Invalid key!\n");
	else
		ofi_idx_remove(&handle->cmap->handles_idx, index);
}

struct util_cmap_handle *ofi_cmap_key2handle(struct util_cmap *cmap, uint64_t key)
{
	struct util_cmap_handle *handle;

	cmap->acquire(&cmap->lock);
	if (!(handle = ofi_idx_lookup(&cmap->handles_idx,
				      ofi_key2idx(&cmap->key_idx, key)))) {
		FI_WARN(cmap->av->prov, FI_LOG_AV, "Invalid key!\n");
	} else {
		if (handle->key != key) {
			FI_WARN(cmap->av->prov, FI_LOG_AV,
				"handle->key not matching given key\n");
			handle = NULL;
		}
	}
	cmap->release(&cmap->lock);
	return handle;
}

/* Caller must hold cmap->lock */
static void util_cmap_init_handle(struct util_cmap_handle *handle,
				  struct util_cmap *cmap,
				  enum util_cmap_state state,
				  fi_addr_t fi_addr,
				  struct util_cmap_peer *peer)
{
	handle->cmap = cmap;
	handle->state = state;
	util_cmap_set_key(handle);
	handle->fi_addr = fi_addr;
	handle->peer = peer;
}

static int util_cmap_match_peer(struct dlist_entry *entry, const void *addr)
{
	struct util_cmap_peer *peer;

	peer = container_of(entry, struct util_cmap_peer, entry);
	return !memcmp(peer->addr, addr, peer->handle->cmap->av->addrlen);
}

/* Caller must hold cmap->lock */
static int util_cmap_del_handle(struct util_cmap_handle *handle)
{
	struct util_cmap *cmap = handle->cmap;
	int ret;

	FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
	       "Deleting connection handle: %p\n", handle);
	if (handle->peer) {
		dlist_remove(&handle->peer->entry);
		free(handle->peer);
		handle->peer = NULL;
	} else {
		cmap->handles_av[handle->fi_addr] = 0;
	}
	util_cmap_clear_key(handle);

	handle->state = CMAP_SHUTDOWN;
	/* Signal CM thread to delete the handle. This is required
	 * so that the CM thread handles any pending events for this
	 * ep correctly. Handle would be freed finally after processing the
	 * events */
	ret = cmap->attr.signal(cmap->ep, handle, OFI_CMAP_FREE);
	if (ret) {
		FI_WARN(cmap->av->prov, FI_LOG_FABRIC,
			"Unable to signal CM thread\n");
		return ret;
	}
	return 0;
}

void ofi_cmap_del_handle(struct util_cmap_handle *handle)
{
	struct util_cmap *cmap = handle->cmap;
	cmap->acquire(&cmap->lock);
	util_cmap_del_handle(handle);
	cmap->release(&cmap->lock);
}

static inline int
util_cmap_check_and_realloc_handles_table(struct util_cmap *cmap, fi_addr_t fi_addr)
{
	int ret = FI_SUCCESS;

	if (OFI_LIKELY(fi_addr < cmap->num_allocated))
		return ret;

	while (fi_addr >= cmap->num_allocated) {
		struct util_cmap_handle **handles_av_new =
				realloc(cmap->handles_av,
					(cmap->av->count +
					 cmap->num_allocated) *
					sizeof(*cmap->handles_av));
		if (OFI_LIKELY(*handles_av_new != NULL)) {
			cmap->handles_av = handles_av_new;
			memset(&cmap->handles_av[cmap->num_allocated],
			       0, sizeof(*cmap->handles_av) *
				  cmap->av->count);
			cmap->num_allocated += cmap->av->count;
		} else {
			ret = -FI_ENOMEM;
			break;
		}
	}

	return ret;
}

/* Caller must hold cmap->lock */
int util_cmap_alloc_handle(struct util_cmap *cmap, fi_addr_t fi_addr,
			   enum util_cmap_state state,
			   struct util_cmap_handle **handle)
{
	int ret;

	*handle = cmap->attr.alloc(cmap);
	if (OFI_UNLIKELY(!*handle))
		return -FI_ENOMEM;
	FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL, "Allocated handle: %p for "
	       "fi_addr: %" PRIu64 "\n", *handle, fi_addr);
	ret = util_cmap_check_and_realloc_handles_table(cmap, fi_addr);
	if (OFI_UNLIKELY(ret)) {
		cmap->attr.free(*handle);
		return ret;
	}
	util_cmap_init_handle(*handle, cmap, state, fi_addr, NULL);
	cmap->handles_av[fi_addr] = *handle;
	return 0;
}

/* Caller must hold cmap->lock */
static int util_cmap_alloc_handle_peer(struct util_cmap *cmap, void *addr,
				       enum util_cmap_state state,
				       struct util_cmap_handle **handle)
{
	struct util_cmap_peer *peer;

	peer = calloc(1, sizeof(*peer) + cmap->av->addrlen);
	if (!peer)
		return -FI_ENOMEM;
	*handle = cmap->attr.alloc(cmap);
	if (!*handle) {
		free(peer);
		return -FI_ENOMEM;
	}
	ofi_straddr_dbg(cmap->av->prov, FI_LOG_AV, "Allocated handle for addr",
			addr);
	FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL, "handle: %p\n", *handle);
	util_cmap_init_handle(*handle, cmap, state, FI_ADDR_NOTAVAIL, peer);
	FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL, "Adding handle to peer list\n");
	peer->handle = *handle;
	memcpy(peer->addr, addr, cmap->av->addrlen);
	dlist_insert_tail(&peer->entry, &cmap->peer_list);
	return 0;
}

/* Caller must hold cmap->lock */
static struct util_cmap_handle *
util_cmap_get_handle_peer(struct util_cmap *cmap, const void *addr)
{
	struct util_cmap_peer *peer;
	struct dlist_entry *entry;

	entry = dlist_find_first_match(&cmap->peer_list, util_cmap_match_peer,
				       addr);
	if (!entry)
		return NULL;
	ofi_straddr_dbg(cmap->av->prov, FI_LOG_AV, "handle found in peer list"
			" for addr", addr);
	peer = container_of(entry, struct util_cmap_peer, entry);
	return peer->handle;
}

int ofi_cmap_move_handle_to_peer_list(struct util_cmap *cmap, int index)
{
	struct util_cmap_handle *handle = cmap->handles_av[index];
	int ret = 0;

	cmap->acquire(&cmap->lock);
	if (!handle)
		goto unlock;

	handle->peer = calloc(1, sizeof(*handle->peer) + cmap->av->addrlen);
	if (!handle->peer) {
		ret = -FI_ENOMEM;
		goto unlock;
	}
	handle->peer->handle = handle;
	memcpy(handle->peer->addr, ofi_av_get_addr(cmap->av, index),
	       cmap->av->addrlen);
	dlist_insert_tail(&handle->peer->entry, &cmap->peer_list);
unlock:
	cmap->release(&cmap->lock);
	return ret;
}

/* Caller must hold cmap->lock */
static int util_cmap_move_handle(struct util_cmap_handle *handle,
				 fi_addr_t fi_addr)
{
	int ret;

	dlist_remove(&handle->peer->entry);
	free(handle->peer);
	handle->peer = NULL;
	handle->fi_addr = fi_addr;
	ret = util_cmap_check_and_realloc_handles_table(handle->cmap, fi_addr);
	if (OFI_UNLIKELY(ret))
		return ret;
	handle->cmap->handles_av[fi_addr] = handle;
	return FI_SUCCESS;
}

int ofi_cmap_update(struct util_cmap *cmap, const void *addr, fi_addr_t fi_addr)
{
	struct util_cmap_handle *handle;
	int ret = 0;

	cmap->acquire(&cmap->lock);
	handle = util_cmap_get_handle_peer(cmap, addr);
	if (!handle) {
		ret = util_cmap_alloc_handle(cmap, fi_addr,
					     CMAP_IDLE, &handle);
		cmap->release(&cmap->lock);
		return ret;
	}
	ret = util_cmap_move_handle(handle, fi_addr);
	cmap->release(&cmap->lock);
	if (ret)
		return ret;

	if (cmap->attr.av_updated_handler)
		cmap->attr.av_updated_handler(handle);
	return 0;
}

/* Caller must hold cmap->lock */

void ofi_cmap_process_shutdown(struct util_cmap *cmap,
			       struct util_cmap_handle *handle)
{
	FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
		"Processing shutdown for handle: %p\n", handle);
	cmap->acquire(&cmap->lock);
	if (handle->state > CMAP_SHUTDOWN) {
		FI_WARN(cmap->av->prov, FI_LOG_EP_CTRL,
			"Invalid handle on shutdown event\n");
	} else if (handle->state != CMAP_SHUTDOWN) {
		FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL, "Got remote shutdown\n");
		util_cmap_del_handle(handle);
	} else {
		FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL, "Got local shutdown\n");
	}
	cmap->release(&cmap->lock);
}

/* Caller must hold cmap->lock */
void ofi_cmap_process_conn_notify(struct util_cmap *cmap,
				  struct util_cmap_handle *handle)
{
	FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
	       "Processing connection notification for handle: %p.\n", handle);
	handle->state = CMAP_CONNECTED;
	cmap->attr.connected_handler(handle);
}

/* Caller must hold cmap->lock */
void ofi_cmap_process_connect(struct util_cmap *cmap,
			      struct util_cmap_handle *handle,
			      uint64_t *remote_key)
{
	FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
	       "Processing connect for handle: %p\n", handle);
	handle->state = CMAP_CONNECTED_NOTIFY;
	if (remote_key)
		handle->remote_key = *remote_key;
}

void ofi_cmap_process_reject(struct util_cmap *cmap,
			     struct util_cmap_handle *handle,
			     enum util_cmap_reject_flag cm_reject_flag)
{
	FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
		"Processing reject for handle: %p\n", handle);
	cmap->acquire(&cmap->lock);
	switch (handle->state) {
	case CMAP_CONNREQ_RECV:
	case CMAP_CONNECTED:
	case CMAP_CONNECTED_NOTIFY:
		/* Handle is being re-used for incoming connection request */
		FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
			"Connection handle is being re-used. Close saved connection\n");
		handle->cmap->attr.close_saved_conn(handle);
		break;
	case CMAP_CONNREQ_SENT:
		if (cm_reject_flag == CMAP_REJECT_GENUINE) {
			FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
			       "Deleting connection handle\n");
			util_cmap_del_handle(handle);
		} else {
			FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
			       "Connection handle is being re-used. Close the connection\n");
			handle->cmap->attr.close(handle);
		}
		break;
	case CMAP_SHUTDOWN:
		FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
			"Connection handle already being deleted\n");
		break;
	default:
		FI_WARN(cmap->av->prov, FI_LOG_EP_CTRL, "Invalid cmap state: "
			"%d when receiving connection reject\n", handle->state);
		assert(0);
	}
	cmap->release(&cmap->lock);
}

int ofi_cmap_process_connreq(struct util_cmap *cmap, void *addr,
			     struct util_cmap_handle **handle_ret,
			     enum util_cmap_reject_flag *cm_reject_flag)
{
	struct util_cmap_handle *handle;
	int ret = 0, index, cmp;

	/* Reset flag to initial state */
	*cm_reject_flag = CMAP_REJECT_GENUINE;

	ofi_straddr_dbg(cmap->av->prov, FI_LOG_EP_CTRL,
			"Processing connreq for addr", addr);

	index = ip_av_get_fi_addr(cmap->av, addr);

	cmap->acquire(&cmap->lock);
	if (index < 0)
		handle = util_cmap_get_handle_peer(cmap, addr);
	else
		handle = ofi_cmap_acquire_handle(cmap, (fi_addr_t)index);

	if (!handle) {
		if (index < 0)
			ret = util_cmap_alloc_handle_peer(cmap, addr,
							  CMAP_CONNREQ_RECV,
							  &handle);
		else
			ret = util_cmap_alloc_handle(cmap, (fi_addr_t)index,
						     CMAP_CONNREQ_RECV,
						     &handle);
		if (ret)
			goto unlock;
	}

	switch (handle->state) {
	case CMAP_CONNECTED_NOTIFY:
	case CMAP_CONNECTED:
		FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
			"Connection already present.\n");
		ret = -FI_EALREADY;
		break;
	case CMAP_CONNREQ_SENT:
		ofi_straddr_dbg(cmap->av->prov, FI_LOG_EP_CTRL, "local_name",
				cmap->attr.name);
		ofi_straddr_dbg(cmap->av->prov, FI_LOG_EP_CTRL, "remote_name",
				addr);

		cmp = ofi_addr_cmp(cmap->av->prov, addr, cmap->attr.name);

		if (cmp < 0) {
			FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
				"Remote name lower than local name.\n");
			*cm_reject_flag = CMAP_REJECT_SIMULT_CONN;
			ret = -FI_EALREADY;
			break;
		} else if (cmp > 0) {
			FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
				"Re-using handle: %p to accept remote "
				"connection\n", handle);
			/* Re-use handle. If it receives FI_REJECT the handle
			 * would not be deleted in this state */
			//handle->cmap->attr.close(handle);
			handle->cmap->attr.save_conn(handle);
		} else {
			FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL,
				"Endpoint connects to itself\n");
			ret = util_cmap_alloc_handle_peer(cmap, addr,
							  CMAP_CONNREQ_RECV,
							  &handle);
			if (ret)
				goto unlock;
			assert(index >= 0 && index != FI_ADDR_NOTAVAIL);
			handle->fi_addr = index;
		}
		/* Fall through */
	case CMAP_IDLE:
		handle->state = CMAP_CONNREQ_RECV;
		/* Fall through */
	case CMAP_CONNREQ_RECV:
		*handle_ret = handle;
		break;
	default:
		FI_WARN(cmap->av->prov, FI_LOG_EP_CTRL,
		       "Invalid cmap state\n");
		assert(0);
		ret = -FI_EOPBADSTATE;
	}
unlock:
	cmap->release(&cmap->lock);
	return ret;
}

/* Caller must hold `cmap::lock` */
int ofi_cmap_handle_connect(struct util_cmap *cmap, fi_addr_t fi_addr,
			    struct util_cmap_handle *handle)
{
	int ret;

	if (handle->state == CMAP_CONNECTED_NOTIFY ||
	    handle->state == CMAP_CONNECTED)
		return FI_SUCCESS;

	switch (handle->state) {
	case CMAP_IDLE:
		ret = cmap->attr.connect(cmap->ep, handle,
					 ofi_av_get_addr(cmap->av, fi_addr),
					 cmap->av->addrlen);
		if (ret) {
			util_cmap_del_handle(handle);
			return ret;
		}
		handle->state = CMAP_CONNREQ_SENT;
		ret = -FI_EAGAIN;
		// TODO sleep on event fd instead of busy polling
		break;
	case CMAP_CONNREQ_SENT:
	case CMAP_CONNREQ_RECV:
	case CMAP_ACCEPT:
	case CMAP_SHUTDOWN:
		ret = -FI_EAGAIN;
		break;
	default:
		FI_WARN(cmap->av->prov, FI_LOG_EP_CTRL,
			"Invalid cmap handle state\n");
		assert(0);
		ret = -FI_EOPBADSTATE;
	}
	return ret;
}

int ofi_cmap_get_handle(struct util_cmap *cmap, fi_addr_t fi_addr,
			struct util_cmap_handle **handle_ret)
{
	int ret;

	cmap->acquire(&cmap->lock);
	*handle_ret = ofi_cmap_acquire_handle(cmap, fi_addr);
	if (OFI_UNLIKELY(!*handle_ret)) {
		ret = -FI_EAGAIN;
		goto unlock;
	}
	
	ret = ofi_cmap_handle_connect(cmap, fi_addr, *handle_ret);
unlock:
	cmap->release(&cmap->lock);
	return ret;
}

static int util_cmap_cm_thread_close(struct util_cmap *cmap)
{
	int ret;

	ret = cmap->attr.signal(cmap->ep, NULL, OFI_CMAP_EXIT);
	if (ret) {
		FI_WARN(cmap->av->prov, FI_LOG_FABRIC,
			"Unable to signal CM thread\n");
		return ret;
	}
	/* Release lock so that CM thread could process shutdown events */
	cmap->release(&cmap->lock);
	ret = pthread_join(cmap->cm_thread, NULL);
	cmap->acquire(&cmap->lock);
	if (ret) {
		FI_WARN(cmap->av->prov, FI_LOG_FABRIC,
			"Unable to join CM thread\n");
		return ret;
	}
	return 0;
}

void ofi_cmap_free(struct util_cmap *cmap)
{
	struct util_cmap_peer *peer;
	struct dlist_entry *entry;
	size_t i;

	cmap->acquire(&cmap->lock);
	FI_DBG(cmap->av->prov, FI_LOG_EP_CTRL, "Closing cmap\n");
	for (i = 0; i < cmap->num_allocated; i++) {
		if (cmap->handles_av[i])
			util_cmap_del_handle(cmap->handles_av[i]);
	}
	while(!dlist_empty(&cmap->peer_list)) {
		entry = cmap->peer_list.next;
		peer = container_of(entry, struct util_cmap_peer, entry);
		util_cmap_del_handle(peer->handle);
	}
	util_cmap_cm_thread_close(cmap);
	cmap->release(&cmap->lock);

	/* cleanup function would be used in manual progress mode */
	if (cmap->attr.cleanup) {
		cmap->attr.cleanup(cmap->ep);
	}
	free(cmap->handles_av);
	free(cmap->attr.name);
	ofi_idx_reset(&cmap->handles_idx);
	if (!cmap->attr.serial_access)
		fastlock_destroy(&cmap->lock);
	free(cmap);
}

static int
util_cmap_update_addr(struct util_av *av, void *addr,
		      fi_addr_t fi_addr, void *arg)
{
	struct util_cmap *cmap = (struct util_cmap *)arg;
	return ofi_cmap_update(cmap, addr, fi_addr);
}

int ofi_cmap_bind_to_av(struct util_cmap *cmap, struct util_av *av)
{
	cmap->av = av;
	return ofi_av_elements_iter(av, util_cmap_update_addr, (void *)cmap);
}

int ofi_cmap_alloc(struct util_ep *ep, struct util_cmap_attr *attr)
{
	struct util_cmap *cmap;
	int ret;

	cmap = calloc(1, sizeof *cmap);
	if (!cmap)
		return -FI_ENOMEM;

	cmap->ep = ep;

	cmap->handles_av = calloc(ep->av->count,
				  sizeof(*cmap->handles_av));
	if (!cmap->handles_av) {
		ret = -FI_ENOMEM;
		goto err1;
	}
	cmap->num_allocated = ep->av->count;

	cmap->attr = *attr;
	cmap->attr.name = mem_dup(attr->name, ep->av->addrlen);
	if (!cmap->attr.name) {
		ret = -FI_ENOMEM;
		goto err2;
	}

	memset(&cmap->handles_idx, 0, sizeof(cmap->handles_idx));
	ofi_key_idx_init(&cmap->key_idx, UTIL_CMAP_IDX_BITS);

	dlist_init(&cmap->peer_list);

	if (cmap->attr.serial_access) {
		cmap->acquire = ofi_fastlock_acquire_noop;
		cmap->release = ofi_fastlock_release_noop;
	} else {
		fastlock_init(&cmap->lock);
		cmap->acquire = ofi_fastlock_acquire;
		cmap->release = ofi_fastlock_release;
	}

	ep->cmap = cmap;

	if (pthread_create(&cmap->cm_thread, 0,
			   cmap->attr.cm_thread_func, ep)) {
		FI_WARN(ep->av->prov, FI_LOG_FABRIC,
			"Unable to create cmap thread\n");
		ret = -FI_ENOMEM;
		goto err3;
	}

	assert(ep->av);
	ret = ofi_cmap_bind_to_av(cmap, ep->av);
	if (ret)
		goto err4;

	return FI_SUCCESS;
err4:
	util_cmap_cm_thread_close(cmap);
err3:
	free(cmap->attr.name);
err2:
	free(cmap->handles_av);
err1:
	free(cmap);
	ep->cmap = NULL;
	return ret;
}
